/*
 *    Collection.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package org.greenstone.gsdl3.collection;

import org.greenstone.gsdl3.util.*;
import org.greenstone.gsdl3.core.*;
import org.greenstone.gsdl3.service.*;


// java XML classes we're using
import org.w3c.dom.Document; 
import org.w3c.dom.Node; 
import org.w3c.dom.Element; 
import org.w3c.dom.NodeList; 

import java.io.File;
import java.util.HashMap;

/**
 * Represents a collection in Greenstone. A collection is an extension of
 * a ServiceCluster - it has local data that the services use.
 *
 * @author <a href="mailto:kjdon@cs.waikato.ac.nz">Katherine Don</a>
 * @version $Revision: 3847 $
 * @see ModuleInterface
 */
public class Collection 
    extends ServiceCluster {

    protected XMLTransformer transformer_ = null;
    /** same as setClusterName */
    public void setCollectionName(String name) {
	setClusterName(name);
    }

    public Collection() {
	service_map_ = new HashMap();
	converter_ = new XMLConverter();
	transformer_ = new XMLTransformer();
	doc_ = converter_.newDOM();
	description_ = doc_.createElement(GSXML.COLLECTION_ELEM);
	service_info_ = doc_.createElement(GSXML.SERVICE_ELEM+GSXML.LIST_MODIFIER);
	meta_info_ = doc_.createElement(GSXML.METADATA_ELEM+GSXML.LIST_MODIFIER);
	description_.appendChild(meta_info_);
	description_.appendChild(service_info_);
	
    }
	
    /**
     * Configures the collection.
     *
     * gsdlHome and collectionName must be set before configure is called.
     *
     * the file buildcfg.xml is located in gsdlHome/collect/collectionName
     * collection metadata is obtained, and services loaded.
     *
     * @return true/false on success/fail
     */
    public boolean configure() {
	
	if (site_home_ == null || cluster_name_== null) {
	    System.err.println("site_home and collection_name must be set before configure called!");
	    return false;
	}
	// read the collection build configuration file
	File build_config_file = new File(GSFile.collectionBuildConfigFile(site_home_, cluster_name_));
	File coll_config_file = new File(GSFile.collectionConfigFile(site_home_, cluster_name_));

	if (!coll_config_file.exists()) {
	    System.err.println(coll_config_file+" does not exist");
	    System.err.println("couldn't configure collection: "+cluster_name_);
	    return false;
	}
	
	if (!build_config_file.exists()) {
	    System.err.println(build_config_file+" does not exist");
	    System.err.println("couldn't configure collection: "+cluster_name_);
	    return false;
	}
	
	// get the xml for both files
	Element coll_config_xml = converter_.getDOM(coll_config_file).getDocumentElement();
	Element build_config_xml = converter_.getDOM(build_config_file).getDocumentElement();
	
	// process the metadata
	Element meta_list = (Element)GSXML.getChildByTagName(coll_config_xml, GSXML.METADATA_ELEM+GSXML.LIST_MODIFIER);
	configureMetadata(meta_list);
	meta_list = (Element)GSXML.getChildByTagName(build_config_xml, GSXML.METADATA_ELEM+GSXML.LIST_MODIFIER);
	configureMetadata(meta_list);
	
	// now do the services
	
	Element service_list = (Element)GSXML.getChildByTagName(build_config_xml, GSXML.SERVICE_CLASS_ELEM+GSXML.LIST_MODIFIER);
	configureServiceRack(service_list, coll_config_xml);
	
	//System.out.println("description=");
	//System.out.println(converter_.getString(description_));
	return true;
	
    }

    /** we need to overwrite this, cos in a collection, some of the info for the service racks is in the collectionConfig.xml file - this needs to be added to the xml for the service rack before configure is called */
    protected boolean configureServiceRack(Element service_rack_list, 
					   Element coll_config_xml) {
	
	// create all the services
	NodeList nodes = service_rack_list.getElementsByTagName(GSXML.SERVICE_CLASS_ELEM);
	if (nodes.getLength()==0) {
	    System.err.println("Cluster configuration error: cluster "+cluster_name_+" has no service modules!");
	    return false;
	}

	for(int i=0; i<nodes.getLength(); i++) {
	
	    // the xml request to send to the serviceRack to query what
	    // services it provides
	    Element message = doc_.createElement(GSXML.MESSAGE_ELEM);
	    Element request = doc_.createElement(GSXML.REQUEST_ELEM);
	    message.appendChild(request);
	    request.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_DESCRIBE);
	    request.setAttribute(GSXML.INFO_ATT, GSXML.SERVICE_ELEM+GSXML.LIST_MODIFIER);
	    

	    Element n = (Element)nodes.item(i);
	    String servicetype = n.getAttribute(GSXML.NAME_ATT);
	    
	    addExtraServiceRackInfo(n, coll_config_xml); // add in any display or format stuff from collectionConfig.xml

	    try {
		ServiceRack s = (ServiceRack)Class.forName("org.greenstone.gsdl3.service."+servicetype).newInstance();

		s.setSiteHome(site_home_);
		s.setClusterName(cluster_name_);
		s.setMessageRouter(router_);
		// pass the xml node to the service for configuration
		s.configure(n);
		
		// find out the supported service types for this service module
		Node types = s.process(message);
		NodeList typenodes = ((Element)types).getElementsByTagName(GSXML.SERVICE_ELEM);	   
		
		for (int j=0; j<typenodes.getLength();j++) {
		    String service = ((Element) typenodes.item(j)).getAttribute(GSXML.NAME_ATT);	    
		    service_map_.put(service, s);
		    
		    // also add info to the ServiceInfo XML element
		    service_info_.appendChild(doc_.importNode(typenodes.item(j), true));
		}
	    } catch (Exception e) {
		System.err.println("ServiceCluster Error:  configure exception: couldn't create service module:org.greenstone.gsdl3.service."+servicetype+"\n"+e.getMessage() + e.getClass() );
		e.printStackTrace();
		//return false;
	    }
	    
	}
	// add elements to description
	description_.appendChild(service_info_);
	
	return true;
	
	
    }

    /** the service rack desription in buildConfig.xml is not complete - 
     * any display or format info specified in the collectionConfig.xml file
     * is not copied over. So that info needs to be added to the service rack
     * description before configuring the service. 
     * extra info is possible for index and classifier elements so far */
    protected boolean addExtraServiceRackInfo(Element service_rack_info,
					      Element coll_config_xml) {

	NodeList indexes = service_rack_info.getElementsByTagName(GSXML.INDEX_ELEM);
	
	if (indexes !=null) {
	    Element config_search = (Element)GSXML.getChildByTagName(coll_config_xml, 
							    GSXML.SEARCH_ELEM);
	    addFormatInfo(indexes, GSXML.INDEX_ELEM, config_search);

	}

	NodeList classifiers = service_rack_info.getElementsByTagName(GSXML.CLASSIFIER_ELEM);
	if (classifiers !=null) {
	    Element config_browse = (Element)GSXML.getChildByTagName(coll_config_xml,
							    GSXML.BROWSE_ELEM);

	    addFormatInfo(classifiers, GSXML.CLASSIFIER_ELEM, config_browse);

	}
	return true;
    }

    protected boolean addFormatInfo(NodeList node_list, String node_name,
				    Element extra_info) {

	//Document doc = node_list.item(0).getOwnerDocument();
	for (int i=0; i<node_list.getLength();i++) {
	    Element node = (Element)node_list.item(i);
	    String name = node.getAttribute(GSXML.NAME_ATT);
	    Document doc = node.getOwnerDocument();
	    Element node_extra = GSXML.getNamedElement(extra_info,
						       node_name,
						       GSXML.NAME_ATT,
						       name);
	    if (node_extra == null) {
		System.out.println("haven't found extra info for "+node_name+" named "+name);
	    }
	    // get the display elements if any - displayName
	    NodeList display_names = node_extra.getElementsByTagName(GSXML.DISPLAYNAME_ELEM);
	    if (display_names !=null) {
		Element display = doc.createElement(GSXML.DISPLAY_ELEM);
		for (int j=0; j<display_names.getLength(); j++) {
		    Element e = (Element)display_names.item(j);
		    
		    Element display_name = GSXML.createTextElement(doc, GSXML.DISPLAY_NAME_ELEM, GSXML.getNodeText(e));
		    display_name.setAttribute(GSXML.LANG_ATT, e.getAttribute(GSXML.LANG_ATT));
		    display.appendChild(display_name);
		}
		node.appendChild(display);
	    }
	    
	    // get the format element if any 
	    Element format = (Element)GSXML.getChildByTagName(node_extra, GSXML.FORMAT_ELEM);
	    if (format==null) {
		format = (Element)GSXML.getChildByTagName(extra_info,
							  GSXML.FORMAT_ELEM);
	    }
	    if (format!=null) { // append to index info
		
		// find the transform xsl file
		String transform_file = GSFile.configFileFormatStylesheet();
		if (transform_file.equals("")) {
		    System.out.println("Collection: couldn't find the config file format transform file!!");
		    // just append the original instead 
		    node.appendChild(doc.importNode(format, true));
		} else {
		    
		    Document stylesheet = converter_.getDOM(new File(transform_file));
		    Element new_format = (Element)transformer_.transform(stylesheet, format);
		    
		    node.appendChild(doc.importNode(new_format, true));
		}
	    }
	}
	return true;
    }

    
}




