/*
 *    GS2Retrieve.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package org.greenstone.gsdl3.service;


// Greenstone classes
import org.greenstone.gdbm.*;
import org.greenstone.gsdl3.util.*;

// XML classes
import org.w3c.dom.Element; 
import org.w3c.dom.NodeList;

// General Java classes
import java.util.StringTokenizer;
import java.util.Vector;


/**
 *
 * @author <a href="mailto:kjdon@cs.waikato.ac.nz">Katherine Don</a>
 * @version $Revision: 3866 $
 */

public abstract class GS2Retrieve
    extends ServiceRack {

    // the services on offer
    // these strings must match what is found in the properties file
    protected static final String DOCUMENT_STRUCTURE_RETRIEVE_SERVICE = "DocumentStructureRetrieve";
    protected static final String DOCUMENT_METADATA_RETRIEVE_SERVICE = "DocumentMetadataRetrieve";
    protected static final String DOCUMENT_CONTENT_RETRIEVE_SERVICE = "DocumentContentRetrieve";

    protected GDBMWrapper gdbm_src_ = null;


    /** constructor */
    protected GS2Retrieve()
    {
	gdbm_src_ = new GDBMWrapper();
    }


    /** configure this service */
    public boolean configure(Element info)
    {
	System.out.println("Configuring GS2Retrieve...");

	// set up short_service_info_ - for now just has name and type
	Element dsr_service = doc_.createElement(GSXML.SERVICE_ELEM);
	dsr_service.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
	dsr_service.setAttribute(GSXML.NAME_ATT, DOCUMENT_STRUCTURE_RETRIEVE_SERVICE);
	short_service_info_.appendChild(dsr_service);

	Element dmr_service = doc_.createElement(GSXML.SERVICE_ELEM);
	dmr_service.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
	dmr_service.setAttribute(GSXML.NAME_ATT, DOCUMENT_METADATA_RETRIEVE_SERVICE);
	short_service_info_.appendChild(dmr_service);

	Element dcr_service = doc_.createElement(GSXML.SERVICE_ELEM);
	dcr_service.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
	dcr_service.setAttribute(GSXML.NAME_ATT, DOCUMENT_CONTENT_RETRIEVE_SERVICE);
	short_service_info_.appendChild(dcr_service);

	// set up service_info_map_ - for now, just has the same elements as above
	// should have full details about each service incl params lists etc.
	service_info_map_.put(DOCUMENT_STRUCTURE_RETRIEVE_SERVICE, dsr_service);
	service_info_map_.put(DOCUMENT_METADATA_RETRIEVE_SERVICE, dmr_service);
	service_info_map_.put(DOCUMENT_CONTENT_RETRIEVE_SERVICE, dcr_service);

	// Open GDBM database for querying
	String gdbm_db_file = GSFile.GDBMDatabaseFile(site_home_, cluster_name_);
	if (!gdbm_src_.openDatabase(gdbm_db_file, GDBMWrapper.READER)) {
	    System.err.println("Error: Could not open GDBM database!");
	    return false;
	}

	return true;
    }


    /** creates a display element containing all the text strings needed to display the service page, in the language specified 
     * these retrieval services dont get displayed to the users - they are only used internally by the actions. so this returns an empty display element*/
    protected Element createServiceDisplay(String service, String lang)
    {
	return doc_.createElement(GSXML.DISPLAY_ELEM);
    }


    /** Returns the parent of a specified documentID, or null if none exists */
    protected Element getParent(String doc_id)
    {
	String parent_id = OID.getParent(doc_id);
	if (parent_id == "")
	    return null;

	Element parent = doc_.createElement(GSXML.DOC_NODE_ELEM);
	parent.setAttribute(GSXML.DOC_NODE_ID_ATT, parent_id);
	if (OID.isTop(parent_id))
	    parent.setAttribute(GSXML.DOC_NODE_TYPE_ATT, GSXML.NODE_TYPE_ROOT);
	else
	    parent.setAttribute(GSXML.DOC_NODE_TYPE_ATT, GSXML.NODE_TYPE_INTERIOR);

	return parent;
    }


    /** */
    protected void addDescendants(Element doc, String doc_id, boolean recursive)
    {
	DBInfo info = gdbm_src_.getInfo(doc_id);
	String contains = info.getInfo("contains");

	StringTokenizer st = new StringTokenizer(contains, ";");
	while (st.hasMoreTokens()) {
	    String child_id = st.nextToken().replaceAll("\"", doc_id);
	    Element child = doc_.createElement(GSXML.DOC_NODE_ELEM);
	    doc.appendChild(child);
	    child.setAttribute(GSXML.DOC_NODE_ID_ATT, child_id);

	    // Apply recursively, if desired
	    if (recursive) {
		addDescendants(child, child_id, recursive);
	    }

	    // Record if the node is a leaf node or an internal node
	    if (isLeafNode(child_id))
		child.setAttribute(GSXML.DOC_NODE_TYPE_ATT, GSXML.NODE_TYPE_LEAF);
	    else
		child.setAttribute(GSXML.DOC_NODE_TYPE_ATT, GSXML.NODE_TYPE_INTERIOR);
	}
    }


    /** Returns true if the OID specifies a leaf node, false otherwise
	Note: this makes a request to the GDBM database so it may not be
	      a particularly cheap operation */
    protected boolean isLeafNode(String oid)
    {
	DBInfo info = gdbm_src_.getInfo(oid);
	String children = info.getInfo("contains");
	return (children == "");
    }


    /** Retrieve the structure of a document */
    protected Element processDocumentStructureRetrieve(Element request)
    {
	// Create a new (empty) result message
	Element result = doc_.createElement(GSXML.RESPONSE_ELEM);
	String from = GSPath.appendLink(cluster_name_, DOCUMENT_STRUCTURE_RETRIEVE_SERVICE);
	result.setAttribute(GSXML.FROM_ATT, from);
	result.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_PROCESS);
	//Element result_content = doc_.createElement(GSXML.CONTENT_ELEM);
	//result.appendChild(result_content);

	// Get the parameters of the request
	Element param_list = (Element) GSXML.getChildByTagName(request, GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	if (param_list == null) {
	    System.err.println("Error: DocumentStructureRetrieve request had no paramList.");
	    return result;  // Return the empty result
	}

	// The document structure information desired
	boolean wantAncestors = false;
	boolean wantParent = false;
	boolean wantSiblings = false;
	boolean wantChildren = false;
	boolean wantDescendants = false;

	// Process the request parameters
	Element param = (Element) param_list.getFirstChild();
	while (param != null) {
	    if (!param.getNodeName().equals(GSXML.PARAM_ELEM)) {
		System.err.println("Warning: Non-param in paramList (ignored).");
	    }
	    else {
		// Identify the structure information desired
		if (param.getAttribute(GSXML.NAME_ATT) == "structure") {
		    String structure = GSXML.getValue(param);
		    // System.out.println("Structure desired: " + structure);

		    // This is NOT locale sensitive
		    if (structure.compareToIgnoreCase("ancestors") == 0)
			wantAncestors = true;
		    else if (structure.compareToIgnoreCase("parent") == 0)
			wantParent = true;
		    else if (structure.compareToIgnoreCase("siblings") == 0)
			wantSiblings = true;
		    else if (structure.compareToIgnoreCase("children") == 0)
			wantChildren = true;
		    else if (structure.compareToIgnoreCase("descendants") == 0)
			wantDescendants = true;
		    else
			System.err.println("Warning: Unknown value \"" + structure + "\".");
		}
	    }

	    param = (Element) param.getNextSibling();
	}

	// Make sure there is no repeated information
	if (wantAncestors)
	    wantParent = false;
	if (wantDescendants)
	    wantChildren = false;

//  	// Get the request content
//  	Element content = (Element) GSXML.getChildByTagName(request, GSXML.CONTENT_ELEM);
//  	if (content == null) {
//  	    System.err.println("Error: DocumentStructureRetrieve request had no content.");
//  	    return result;
//  	}
	
	Element query_doc_list = (Element) GSXML.getChildByTagName(request, GSXML.DOC_NODE_ELEM+GSXML.LIST_MODIFIER);
	if (query_doc_list == null) {
	    System.err.println("Error: DocumentStructureRetrieve request specified no doc nodes.\n");
	    return result;
	}

	Element doc_list = doc_.createElement(GSXML.DOC_NODE_ELEM+GSXML.LIST_MODIFIER);
	result.appendChild(doc_list);

	// Get the documents
	String[] doc_ids = GSXML.getAttributeValuesFromList(query_doc_list, 
							    GSXML.DOC_NODE_ID_ATT);
	for (int i = 0; i < doc_ids.length; i++) {
	    String doc_id = doc_ids[i];
	    
	    // Add the document to the list
	    Element doc = doc_.createElement(GSXML.DOC_NODE_ELEM);
	    doc_list.appendChild(doc);
	    doc.setAttribute(GSXML.DOC_NODE_ID_ATT, doc_id);
	    // all structure info goes into a docNodestructure elem
	    Element structure_elem = doc_.createElement(GSXML.DOC_NODE_STRUCTURE_ELEM);
	    doc.appendChild(structure_elem);
	    // Add the requested structure information
	    Element current = doc_.createElement(GSXML.DOC_NODE_ELEM);
	    current.setAttribute(GSXML.DOC_NODE_ID_ATT, doc_id);
	    if (OID.isTop(doc_id))
		current.setAttribute(GSXML.DOC_NODE_TYPE_ATT, GSXML.NODE_TYPE_ROOT);
	    else if (isLeafNode(doc_id))
		current.setAttribute(GSXML.DOC_NODE_TYPE_ATT, GSXML.NODE_TYPE_LEAF);
	    else
		current.setAttribute(GSXML.DOC_NODE_TYPE_ATT, GSXML.NODE_TYPE_INTERIOR);

	    // Ancestors: continually add parent nodes until the root is reached
	    if (wantAncestors) {
		String current_id = doc_id;
		Element current_node = current;
		while (true) {
		    Element parent = getParent(current_id);
		    if (parent == null)
			break;

		    parent.appendChild(current_node);
		    current_node = parent;
		    current_id = OID.getParent(current_id);
		}
	    }
	    // Parent: get the parent of the selected node
	    if (wantParent) {
		Element parent = getParent(doc_id);
		if (parent != null) {
		    parent.appendChild(current);
		}
	    }

	    // Siblings: get the other descendants of the selected node's parent
	    if (wantSiblings) {
		// !! TO DO !!
	    }

	    // Children: get the descendants, but only one level deep
	    if (wantChildren)
		addDescendants(current, doc_id, false);
	    // Descendants: recursively get every descendant of the selected node
	    if (wantDescendants)
		addDescendants(current, doc_id, true);

	    // Add the root of the structure information to the document
	    Element root = current;
	    while (true) {
		if (root.getParentNode() == null)
		    break;

		root = (Element) root.getParentNode();
	    }
	    structure_elem.appendChild(root);
	}

	return result;
    }


    /** Retrieve metadata associated with a document */
    protected Element processDocumentMetadataRetrieve(Element request)
    {
	// Create a new (empty) result message
	Element result = doc_.createElement(GSXML.RESPONSE_ELEM);
	String from = GSPath.appendLink(cluster_name_, DOCUMENT_METADATA_RETRIEVE_SERVICE);
	result.setAttribute(GSXML.FROM_ATT, from);
	result.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_PROCESS);
	//Element result_content = doc_.createElement(GSXML.CONTENT_ELEM);
	//result.appendChild(result_content);

	// Get the parameters of the request
	Element param_list = (Element) GSXML.getChildByTagName(request, GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	if (param_list == null) {
	    System.err.println("GS2Retrieve, DocumentMetadataRetrieve Error: missing paramList.\n");
	    return result;  // Return the empty result
	}

	// The metadata information required
	Vector metadata_list = new Vector();

	// Process the request parameters
	Element param = (Element) param_list.getFirstChild();
	while (param != null) {
	    if (!param.getNodeName().equals(GSXML.PARAM_ELEM)) {
		System.err.println("Warning: Non-param in paramList (ignored).");
	    }
	    else {
		// Identify the metadata information desired
		if (param.getAttribute(GSXML.NAME_ATT) == "metadata") {
		    String metadata = GSXML.getValue(param);
		    System.out.println("Metadata desired: " + metadata);
		    metadata_list.add(metadata);
		}
	    }

	    param = (Element) param.getNextSibling();
	}

	// Get the request content
//  	Element content = (Element) GSXML.getChildByTagName(request, GSXML.CONTENT_ELEM);
//  	if (content == null) {
//  	    System.err.println("Error: DocumentMetadataRetrieve request had no content.\n");
//  	    return result;
//  	}

	Element doc_list = doc_.createElement(GSXML.DOC_NODE_ELEM+GSXML.LIST_MODIFIER);
	//result_content.appendChild(doc_list);
	result.appendChild(doc_list);

	// Get the documents
	//Element request_doc_list = (Element) GSXML.getChildByTagName(content, GSXML.DOCUMENT_ELEM+GSXML.LIST_MODIFIER);
	Element request_doc_list = (Element) GSXML.getChildByTagName(request, GSXML.DOC_NODE_ELEM+GSXML.LIST_MODIFIER);
	if (request_doc_list == null) {
	    System.err.println("Error: DocumentMetadataRetrieve request had no documentList.\n");
 	    return result;
 	}
	NodeList request_docs = request_doc_list.getChildNodes();
	for (int i = 0; i < request_docs.getLength(); i++) {
	    Element request_doc = (Element) request_docs.item(i);
	    String doc_id = request_doc.getAttribute(GSXML.DOC_NODE_ID_ATT);
	    System.out.println("Retrieving metadata for " + doc_id + "...");

	    // Add the document to the list
	    Element doc_node = doc_.createElement(GSXML.DOC_NODE_ELEM);
	    doc_list.appendChild(doc_node);
	    doc_node.setAttribute(GSXML.DOC_NODE_ID_ATT, doc_id);
	    doc_node.setAttribute(GSXML.DOC_NODE_TYPE_ATT,
				  request_doc.getAttribute(GSXML.DOC_NODE_TYPE_ATT));

	    // Add the requested metadata information
	    Element doc_meta_list = doc_.createElement(GSXML.METADATA_ELEM+GSXML.LIST_MODIFIER);
	    doc_node.appendChild(doc_meta_list);
	    DBInfo info = gdbm_src_.getInfo(doc_id);
	    for (int m = 0; m < metadata_list.size(); m++) {
		String metadata = (String) metadata_list.get(m);
		String value = info.getInfo(metadata);
		GSXML.addMetadata(doc_, doc_meta_list, metadata, value);
	    }
	}

	return result;
    }


    /** Retrieve the content of a document - implemented by concrete subclasses */
    protected abstract Element processDocumentContentRetrieve(Element request);
}   
